package core

import (
	"witness/quest"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
)

func (w *Witness) LoadPendingProofs(startTimestamp, endTimestamp int64) {
	defaultLimit := 100
	lastTaskID := ""
	var dbProofs []*quest.ProofModel
	var err error
	for {
		dbProofs, err = w.q.GetProofs(startTimestamp, endTimestamp, lastTaskID, defaultLimit)
		if err != nil {
			log.WithError(err).Error("failed to get proofs")
			return
		}
		log.WithField("proof count", len(dbProofs)).Debug("load pending proofs")

		if len(dbProofs) == 0 {
			return
		}

		for _, dbProof := range dbProofs {
			miner, proof := w.VerifyProof(dbProof)
			if proof != nil {
				// w.AddPendingProof(miner, proof)
			}
			_ = miner
			lastTaskID = dbProof.TaskId
		}
	}
}

func (w *Witness) VerifyProof(dbProof *quest.ProofModel) (miner common.Address, proof *witnessv1.ValidatedProof) {
	if dbProof.TaskWorkload == 0 {
		return
	}
	containerPubKey, err := crypto.SigToPub(dbProof.GenerateHashPayload(), common.Hex2Bytes(dbProof.TaskContainerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify container signature")
		return
	}
	containerSigner := crypto.PubkeyToAddress(*containerPubKey)
	verified := w.IsContainerAddress(containerSigner)
	if !verified {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "signer": containerSigner.Hex()}).Error("invalid container signature")
		return
	}

	minerPubKey, err := crypto.SigToPub(dbProof.GenerateHashPayload(), common.Hex2Bytes(dbProof.TaskMinerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify miner signature")
		return
	}
	workerAddress := crypto.PubkeyToAddress(*minerPubKey)
	if workerAddress.Hex() != common.HexToAddress(dbProof.TaskWorkerAccount).Hex() {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "miner": dbProof.TaskWorkerAccount, "signer": workerAddress.Hex()}).Error("invalid miner address")
		return
	}

	nmPubKey, err := crypto.SigToPub(dbProof.GenerateNMHashPayload(), common.Hex2Bytes(dbProof.TaskManagerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify manager signature")
		return
	}

	nmAddress := crypto.PubkeyToAddress(*nmPubKey)
	verified = w.IsNMAddress(nmAddress)
	if !verified {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "signer": nmAddress.Hex()}).Error("invalid manager signature")
		return
	}

	proof = &witnessv1.ValidatedProof{
		Workload:  dbProof.TaskWorkload,
		Timestamp: uint64(dbProof.TaskFinishTimestamp),
	}
	return common.HexToAddress(dbProof.TaskProfitAccount), proof
}
