package core

import (
	"encoding/json"
	"fmt"
	"math/big"
	"math/rand"
	"os"
	"sort"
	"sync"
	"time"
	"witness/conf"
	"witness/quest"
	"witness/tree"
	"witness/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethdb/leveldb"
	"github.com/ethereum/go-ethereum/trie"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
	"github.com/syndtr/goleveldb/leveldb/errors"
)

type Witness struct {
	lvdb               *leveldb.Database
	db                 *trie.Database
	state              *StateDB
	pendingProof       map[int64]map[common.Address]*witnessv1.ValidatedProof
	pendingWorkload    uint64
	mtTreeCache        map[string]*tree.MerkleTree
	date               string
	rpc                *ChainRPC
	containerAddresses []common.Address
	nmAddresses        []common.Address
	q                  *quest.Quest

	cfg *conf.Config

	sync.Mutex
}

func RunWitness(q *quest.Quest, cfg *conf.Config) *Witness {
	_rpc := newChain(cfg.ChainRPC, cfg.PrivateKey, cfg.StoreContract, cfg.WitnessContract)
	diskDB, err := leveldb.New(fmt.Sprintf("%s/db", cfg.DataDir), 128, 1024, "", false)
	if err != nil {
		panic(err)
	}
	var lastDay string
	ok, err := diskDB.Has([]byte("lastday"))
	if err != nil {
		panic(err)
	}
	var stateRoot common.Hash
	if ok {
		val, _ := diskDB.Get([]byte("lastday"))
		log.WithField("last day", string(val)).Info("diskdb get last day")
		lastDay = string(val)
		root, _ := diskDB.Get([]byte(fmt.Sprintf("sroot:%s", string(val))))
		stateRoot = common.BytesToHash(root)
		log.WithFields(log.Fields{
			"last day":   lastDay,
			"state root": stateRoot.Hex(),
		}).Info("load state from diskdb")
	}
	stateDB, err := NewStateDB(diskDB, stateRoot)
	if err != nil {
		panic(err)
	}
	w := &Witness{
		lvdb:         diskDB,
		state:        stateDB,
		pendingProof: make(map[int64]map[common.Address]*witnessv1.ValidatedProof),
		mtTreeCache:  make(map[string]*tree.MerkleTree),
		rpc:          _rpc,
		date:         lastDay,
		cfg:          cfg,
		q:            q,
	}
	w.LoadMerkleProof(lastDay)
	w.LoadPendingProof()
	go w.StorePendingProofJob()
	go w.UpdateAddress()
	<-time.After(time.Second * 3)
	go w.Ticker()
	// go w.Mock()
	// go w.ProcessDay()

	return w
}

func (w *Witness) AddPendingProof(miner common.Address, proof *witnessv1.ValidatedProof) {
	w.Lock()
	defer w.Unlock()

	var userTodayWorkload uint64
	currentPendingProof, ok := w.pendingProof[w.todayTimestamp()][miner]
	if ok {
		userTodayWorkload = currentPendingProof.Workload
	}

	w.pendingWorkload += proof.Workload
	if w.pendingProof[w.todayTimestamp()] == nil {
		w.pendingProof[w.todayTimestamp()] = make(map[common.Address]*witnessv1.ValidatedProof)
	}
	w.pendingProof[w.todayTimestamp()][miner] = &witnessv1.ValidatedProof{Workload: userTodayWorkload + proof.Workload}

	log.WithFields(log.Fields{
		"miner":            miner.Hex(),
		"current_workload": userTodayWorkload + proof.Workload,
		"workload":         proof.Workload,
		"global_workload":  w.pendingWorkload,
	}).Debug("add validated proof")
}

func (w *Witness) ProcessDay() {
	log.Debugln("process day", w.yesterdayTimestamp(), w.todayTimestamp())
	w.LoadPendingProofs(w.yesterdayTimestamp(), w.todayTimestamp())
	w.date = w.yesterdayString()
	dayProof := w.Commit()
	mstRoot, _, err := w.CommitMST(dayProof)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle sum tree")
		return
	}
	objects := w.state.IterAllObject()
	mtRoot, err := w.CommitMT(objects)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle tree")
		return
	}
	err = w.lvdb.Put([]byte("lastday"), []byte(w.date))
	if err != nil {
		log.WithError(err).Error("db failed to update last day")
		return
	}
	txHash, err := w.rpc.SubmitProofs(w.date, mstRoot, mtRoot)
	if err != nil {
		log.WithError(err).Error("submit proofs")
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("txid:%s", w.date)), txHash.Bytes())
	if err != nil {
		log.WithError(err).Error("db failed to save txid")
		return
	}
	log.WithField("tx_hash", txHash.Hex()).Info("submit proofs")
	log.Info("process day done")
}

func (w *Witness) TotalWorkload() {

}

func (w *Witness) Commit() (dayProofs map[common.Address]*witnessv1.ValidatedProof) {
	st := time.Now()
	proof, totalWorkload := w.RefreshPendingProof()
	balancePerWorkload, err := w.rpc.GetWorkloadThreshold(totalWorkload)

	for miner, proof := range proof {
		balance := big.NewInt(0).Mul(balancePerWorkload, big.NewInt(int64(proof.Workload)))
		proof.Balance = balance.String()
		err := w.SealProof(miner, proof)
		if err != nil {
			log.WithError(err).Error("failed to seal proof")
			return
		}
	}
	root, err := w.state.Commit()
	if err != nil {
		return
	}
	log.WithFields(log.Fields{"root": root.Hex(), "cost": time.Since(st).String()}).Info("commit state root")
	err = w.lvdb.Put([]byte(fmt.Sprintf("sroot:%s", w.date)), root.Bytes())
	if err != nil {
		log.WithError(err).Error("failed to update state root")
		return
	}
	return proof
}

func (w *Witness) SealProof(miner common.Address, proof *witnessv1.ValidatedProof) (err error) {
	minerObject := w.state.GetMinerObject(miner)
	oldBalance := minerObject.Balance
	minerObject.Balance = util.MustAddStringToBigInt(oldBalance, proof.Balance).String()
	err = w.state.UpdateMinerObject(miner, minerObject)
	if err != nil {
		return err
	}
	log.WithFields(log.Fields{
		"miner":       miner.Hex(),
		"old_balance": oldBalance,
		"new_balance": minerObject.Balance,
	}).Info("seal proof")
	return err
}

func (w *Witness) RefreshPendingProof() (proof map[common.Address]*witnessv1.ValidatedProof, totalWorkload uint64) {
	w.Lock()
	defer w.Unlock()
	proof = make(map[common.Address]*witnessv1.ValidatedProof)
	for timestamp, pendingProofMap := range w.pendingProof {
		if timestamp != w.todayTimestamp() {
			continue
		}
		for miner, pendingProof := range pendingProofMap {
			proof[miner] = pendingProof
			totalWorkload += pendingProof.Workload
		}
	}
	w.pendingProof[w.yesterdayTimestamp()] = make(map[common.Address]*witnessv1.ValidatedProof)
	return
}

// CommitMST commit workload of per day
func (w *Witness) CommitMST(proofMap map[common.Address]*witnessv1.ValidatedProof) (root common.Hash, sum *big.Int, err error) {
	if len(proofMap) == 0 {
		return common.Hash{}, big.NewInt(0), nil
	}
	rawKeys := make([]common.Address, 0)
	keys := make([][32]byte, len(proofMap))
	dbKey := make([]byte, 0)
	dbVal := make([]byte, 0)
	vals := make([]*big.Int, len(proofMap))
	for key := range proofMap {
		rawKeys = append(rawKeys, key)
	}
	sort.Sort(util.AddressSlice(rawKeys))
	for i, key := range rawKeys {
		keys[i] = crypto.Keccak256Hash(key.Bytes())
		dbKey = append(dbKey, keys[i][:]...)
		vals[i] = new(big.Int).SetUint64(proofMap[key].Workload)
		dbVal = append(append(dbVal, vals[i].Bytes()...), []byte(":")...)
	}

	st := time.Now()

	nodes, sums := tree.ComputeMST(keys, vals)

	err = w.lvdb.Put([]byte(fmt.Sprintf("mstroot:%s", w.date)), nodes[0][:])
	if err != nil {
		log.Error(err)
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("mstsum:%s", w.date)), sums[0].Bytes())
	if err != nil {
		log.Error(err)
		return
	}

	err = w.lvdb.Put([]byte(fmt.Sprintf("mstk:%s", w.date)), dbKey)
	if err != nil {
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("mstv:%s", w.date)), dbVal)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"root": common.Hash(nodes[0]).Hex(),
		"sum":  sums[0].String(),
		"cost": time.Since(st).String(),
	}).Info("commit MST root")

	return nodes[0], sums[0], nil
}

// CommitMT commit all workload
func (w *Witness) CommitMT(objects []*witnessv1.MinerObject) (root common.Hash, err error) {
	if len(objects) == 0 {
		return common.Hash{}, nil
	}
	merkleProofs := make(tree.Proofs, 0)
	dbProofs := make([]byte, 0)
	for _, object := range objects {
		bigBalance, _ := new(big.Int).SetString(object.Balance, 10)
		payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(bigBalance.Bytes(), 32)...)
		_proof := crypto.Keccak256Hash(payload)
		merkleProofs = append(merkleProofs, _proof)
		dbProofs = append(dbProofs, _proof[:]...)
	}

	mtTree, err := tree.NewMerkleTree(merkleProofs)
	if err != nil {
		return
	}
	root = mtTree.GetRoot()

	st := time.Now()
	err = w.lvdb.Put([]byte(fmt.Sprintf("mtroot:%s", w.date)), root.Bytes())
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k": fmt.Sprintf("mtroot:%s", w.date),
		"v": root.String(),
	}).Debug()

	err = w.lvdb.Put([]byte(fmt.Sprintf("mtk:%s", w.date)), dbProofs)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k":        fmt.Sprintf("mtk:%s", w.date),
		"v_length": len(dbProofs),
	}).Debug()
	w.mtTreeCache[w.date] = mtTree

	log.WithFields(log.Fields{
		"root": root.Hex(),
		"cost": time.Since(st).String(),
	}).Info("commit MT root")

	return
}

func (w *Witness) Ticker() {
	executionTime := time.Date(time.Now().Year(), time.Now().Month(), time.Now().Day(), 0, 0, w.cfg.CommitTime, 0, time.UTC).Add(w.duration())
	// executionTime := time.Now().Add(w.duration())
	waitTime := executionTime.Sub(time.Now())
	timer := time.NewTimer(waitTime)
	log.WithField("wait_time", waitTime.String()).Info("prepare commit task")
	for {
		<-timer.C
		w.ProcessDay()
		executionTime = executionTime.Add(w.duration())
		waitTime = executionTime.Sub(time.Now())
		timer.Reset(waitTime)
		log.WithField("wait time", waitTime.String()).Info("prepare commit task")
	}
}

func (w *Witness) Mock() {
	addresses := make([]common.Address, 10)
	rand.Seed(time.Now().Unix())
	for i := 0; i < 10; i++ {
		b20 := make([]byte, 20)
		rand.Read(b20)
		addresses[i] = common.BytesToAddress(b20)
	}

	for {
		time.Sleep(time.Second)
		w.AddPendingProof(addresses[rand.Intn(10)], &witnessv1.ValidatedProof{
			Workload:  uint64((rand.Int63n(9) + 1) * 100),
			Timestamp: uint64(time.Now().Unix() - 1),
		})
	}
}

func (w *Witness) UpdateAddress() {
	ticker := time.NewTicker(time.Minute * 10)
	defer ticker.Stop()
	log.Info("start update address task")
	for {
		addrs, err := w.rpc.GetContainerAddresses()
		if err != nil {
			log.WithError(err).Error("failed to get container addresses")
		} else {
			w.containerAddresses = addrs
		}
		addrs, err = w.rpc.GetNMAddresses()
		if err != nil {
			log.WithError(err).Error("failed to get NM addresses")
		} else {
			w.nmAddresses = addrs
		}
		log.WithFields(log.Fields{
			"container_count": len(w.containerAddresses),
			"nm_count":        len(w.nmAddresses),
		}).Info("store contract update address")
		<-ticker.C
	}
}

func (w *Witness) IsContainerAddress(address common.Address) bool {
	for _, addr := range w.containerAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) IsNMAddress(address common.Address) bool {
	for _, addr := range w.nmAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) GetMerkleProof(address common.Address, date string) (balance string, proofs []common.Hash) {
	if date == "" {
		date = w.date
	}
	w.Lock()
	cacheTree, ok := w.mtTreeCache[date]
	w.Unlock()
	if !ok {
		if ok = w.LoadMerkleProof(date); !ok {
			log.WithFields(log.Fields{
				"date": date,
			}).Error("load merkle proof empty")
			return "0", nil
		}
	}
	cacheTree = w.mtTreeCache[date]
	dateStateRootKey := fmt.Sprintf("sroot:%s", date)
	dateStateRoot, err := w.lvdb.Get([]byte(dateStateRootKey))
	if err != nil {
		log.WithFields(log.Fields{
			"key": dateStateRootKey,
			"err": err.Error(),
		}).Error("failed to get state root")
		return "0", nil
	}
	var sdb *StateDB
	if date == w.date {
		sdb = w.state
	} else {
		sdb, err = NewStateDB(w.lvdb, common.BytesToHash(dateStateRoot))
		if err != nil {
			log.WithError(err).Error("failed to create state db")
			return "0", nil
		}
	}
	object := sdb.GetMinerObject(address)
	if object == nil {
		return "0", nil
	}
	bigBalance, _ := new(big.Int).SetString(object.Balance, 10)
	payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(bigBalance.Bytes(), 32)...)
	leaf := crypto.Keccak256Hash(payload)
	proofs, err = cacheTree.GetProof(leaf)
	if err != nil {
		log.WithError(err).Error("failed to get merkle proof")
		return "0", make([]common.Hash, 0)
	}
	return object.Balance, proofs
}

func (w *Witness) GetDailyMerkleProofs(date string, depth int, rootHash common.Hash) (proofs [][]common.Hash) {
	if date == "" {
		date = w.date
	}
	w.Lock()
	cacheTree, ok := w.mtTreeCache[date]
	w.Unlock()
	if !ok {
		if ok = w.LoadMerkleProof(date); !ok {
			log.WithFields(log.Fields{
				"date": date,
			}).Error("load merkle proof empty")
			return nil
		}
	}
	cacheTree = w.mtTreeCache[date]

	rootNode := cacheTree.GetRootNode()
	if rootHash.Hex() != rootNode.Hash.Hex() {
		rootNode = cacheTree.FindNode(rootHash)
	}
	return tree.Traversal(rootNode, depth)
}

func (w *Witness) GetPendingWorkload(address common.Address) (workload, globalWorkload uint64) {
	w.Lock()
	proof, ok := w.pendingProof[w.todayTimestamp()][address]
	w.Unlock()
	if !ok {
		return 0, w.pendingWorkload
	}
	return proof.Workload, w.pendingWorkload
}

func (w *Witness) LoadMerkleProof(date string) (ok bool) {
	if date == "" {
		return false
	}
	merkleTreeKey := fmt.Sprintf("mtk:%s", date)
	log.Warnln("load merkle proof", merkleTreeKey)
	data, err := w.lvdb.Get([]byte(merkleTreeKey))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle proof")
		return
	}

	proofs := make([]common.Hash, len(data)/32)
	for i := 0; i < len(data)/32; i++ {
		copy(proofs[i][:], data[i*32:(i+1)*32])
	}

	mTree, err := tree.NewMerkleTree(proofs)
	if err != nil {
		log.WithError(err).Error("failed to load merkle proof")
		return
	}
	w.Lock()
	w.mtTreeCache[date] = mTree
	w.Unlock()
	log.WithFields(log.Fields{"date": date, "root": mTree.GetRoot().Hex()}).Info("load merkle tree")
	return true
}

func (w *Witness) StorePendingProofJob() {
	ticker := time.NewTicker(time.Second * 30)
	defer ticker.Stop()
	log.Info("start store pending proof task")
	for {
		<-ticker.C
		w.StorePendingProof()
	}
}

func (w *Witness) StorePendingProof() {
	w.Lock()
	proofs := w.pendingProof[w.todayTimestamp()]
	w.Unlock()
	filename := fmt.Sprintf("%s/pending_%s", w.cfg.DataDir, w.todayString())
	exist := true
	_, err := os.Stat(filename)
	if err != nil {
		if os.IsNotExist(err) {
			exist = false
		} else {
			log.WithError(err).Error("failed to store pending proof")
			return
		}
	}
	if exist {
		err = os.Remove(filename)
		if err != nil {
			log.WithError(err).Error("failed to store pending proof")
			return
		}
	}
	st := time.Now()
	data, _ := json.Marshal(proofs)

	err = os.WriteFile(filename, data, 0644)
	if err != nil {
		log.WithError(err).Error("failed to store pending proof")
	}
	log.WithFields(log.Fields{
		"filename": filename,
		"cost":     fmt.Sprintf("%d ms", time.Since(st).Milliseconds()),
	}).Info("store pending proof success")
}

func (w *Witness) LoadPendingProof() {
	proofs := make(map[common.Address]*witnessv1.ValidatedProof)
	filename := fmt.Sprintf("%s/pending_%s", w.cfg.DataDir, w.todayString())
	_, err := os.Stat(filename)
	if err != nil {
		if os.IsNotExist(err) {
			log.WithError(err).Debug("no pending proof")
			return
		}
		log.WithError(err).Error("failed to load pending proof, file status")
		return
	}
	st := time.Now()
	data, err := os.ReadFile(filename)
	if err != nil {
		log.WithError(err).Error("failed to load pending proof")
		return
	}

	err = json.Unmarshal(data, &proofs)
	if err != nil {
		log.WithError(err).Error("failed to load pending proof, json unmarshal error")
		return
	}

	w.Lock()
	w.pendingProof[w.todayTimestamp()] = proofs
	w.Unlock()

	log.WithFields(log.Fields{
		"filename": filename,
		"cost":     fmt.Sprintf("%d ms", time.Since(st).Milliseconds()),
	}).Info("load pending proof success")
}

func (w *Witness) todayString() string {
	return time.Unix(w.todayTimestamp(), 0).Format("2006-01-02")
}

func (w *Witness) yesterdayString() string {
	return time.Unix(w.yesterdayTimestamp(), 0).Format("2006-01-02")
}

func (w *Witness) todayTimestamp() int64 {
	now := time.Now().UTC()
	// today := time.Date(now.Year(), now.Month(), now.Day(), 0, 0, 0, 0, time.UTC)
	today := time.Date(now.Year(), now.Month(), 6, 0, 0, 0, 0, time.UTC)
	return today.Unix()
}

func (w *Witness) yesterdayTimestamp() int64 {
	return w.todayTimestamp() - 86400
}

func (w *Witness) dateToTimestamp(date string) int64 {
	t, _ := time.ParseInLocation("2006-01-02", date, time.UTC)
	return t.Unix()
}

func (w *Witness) duration() time.Duration {
	return 3600 * 24 * time.Second
}
